#include "dilations.h"

// ordering used is x1 x1' x2 x2' ... xn xn'

// maximum allocation
// 40 million nodes = 800 megabytes
int maxnodenum = 20 * 1000 * 1000;
// BDD lib config variables
int nodenum = 1000 * 1000;
int cacheratio = 64;

// how many prop vars in a world
unsigned propvars = 3;

// how many dilations have happened?
// used for benchmarking
unsigned dilations = 0;

// how many conjunctions have been tested for?
// used for benchmarking
unsigned conjunctions = 0;

// // propositional or first-order printing mode
// bool first_order = false;
// // first-order domain size 
// unsigned dsize = 1;
// // first-order predicate arity
// unsigned arity = 1;
// // number of first-order predicates
// unsigned predicates = 2;

Signature signature;

// boolean flag for termination after timer expiration
volatile bool timedOut = false;
// boolean flag for termination after memory exhaustion
volatile bool outOfMem = false;

// set of prop vars in a world
// bdd varset;

// set of pairs of vars to substitute
// bddPair * substPairs = NULL;

// relation representing worlds at distance one or zero
// bdd d10;

// initialise seed from current time
unsigned getSeed(void) {
	time_t t1;
	(void) time(&t1);
	return (unsigned) t1;
}

static void printPred(char *varset, int v) {
	if(varset[v] < 0) {
		std::cout << '?';
	} else if (varset[v]==0) {
		std::cout << '-';
	} else {
		std::cout << ' ';
	}

	unsigned pred = v % signature.predicates;
	v -= pred;
	v /= signature.predicates;
	
	std::cout << (char)('P' + pred) << '(';
	
	for(unsigned i=0; i<signature.arity-1; ++i) {
		unsigned tmp = v % signature.dsize;
		v -= tmp;
		v /= signature.dsize;
		std::cout << (char)('a' + tmp) << ',';
	}
	
	std::cout << (char)('a' + v) << ')';
}

// handler for enumerating models that satisfy a BDD
void allsatPrintHandler(char *varset, int size) {
	if(!signature.first_order) {
		int v;
		
		std::cout << "( ";
		for(v=0; v<size-1; v++) {
			if(varset[v] < 0) {
				std::cout << "X, ";
			} else {
				std::cout << (char) ('0' + varset[v]) << ", ";
			}
		}
		if(varset[v] < 0) {
			std::cout << "X";
		} else {
			std::cout << (char) ('0' + varset[v]);
		}
		std::cout << " )," << std::endl;
	} else {
		std::cout << "( ";
		
		for(int v=0; v<size-1; ++v) {
			printPred(varset, v);
			std::cout << ' ';
		}
		printPred(varset, size-1);
		std::cout << " )\n";
	}
}

// handler for pretty-printing variables
// when outputing a bdd as a dot file
static void printhandler(std::ostream &o, int var) {
	if(!signature.first_order) {
		o << 'X' << var;
		
	} else {
/*		unsigned relem = var % dsize;
		var -= relem;
		var /= dsize;
		unsigned lelem = var % dsize;
		var -= lelem;
		var /= dsize;
		unsigned pred = var;
		o << (char)('P' + pred) << '(' << (char)('a'+lelem) << ',' << (char)('a'+relem) << ')';*/
	}
}

// custom error handler should allow normal exiting from
// an out-of-memory BDD operation, and abort in any other error
static void errorhandler(int errcode) {
	// if out of memory then set terminate flag and return normally
	if(errcode==BDD_MEMORY) {
		outOfMem = true;
	} else {
		std::cerr << "BDD error: " << bdd_errstring(errcode) << ", aborting.\n";
		abort();
	}
}

// builds bdd for worlds of (dalal) distance one
// static void build_d10(void) {
// 	d10 = bddtrue;
// 	// first case is when distance is zero
// 	// ie when x_i <=> x_i'
// 	for(unsigned i=0; i<propvars; i++) {
// 		d10 &= bdd_biimp( bdd_ithvar(2*i), bdd_ithvar(2*i+1) );
// 	}	
// 	
// 	// now build the disjunctions accounting for
// 	// distance one
// 	for(unsigned i=0; i<propvars; i++) {
// 		bdd term = bddtrue;
// 		
// 		for(unsigned j=0; j<propvars; j++) {
// 			if(i==j) {
// 				term &= bdd_xor(bdd_ithvar(2*j), bdd_ithvar(2*j+1));
// 			} else {
// 				term &= bdd_biimp(bdd_ithvar(2*j), bdd_ithvar(2*j+1));
// 			}
// 		}
// 		
// 		d10 |= term;
// 	}
// }

// initialises bdd package and basic variables
void init(void) {
	// reset flags
	timedOut = false;
	outOfMem = false;
	
	// make sure package is not already initialised
	assert( !bdd_isrunning() );
	
	// initialise
	assert( bdd_init(nodenum, nodenum/cacheratio)==0 );
	assert( bdd_setcacheratio( cacheratio ) >= 0 );
	
	// buddy finds the smallest prime larger than nodenum
	// so after initialising, we may have to reset maxnodenum to reflect that
	maxnodenum = std::max(bdd_getallocnum()+1, maxnodenum);
	assert( bdd_setmaxnodenum( maxnodenum ) >= 0 );
	assert( bdd_setmaxincrease( maxnodenum ) >= 0 );
	
	// will need two copies of atoms
	//assert( bdd_setvarnum(2 * propvars)==0 );
	assert( bdd_setvarnum(propvars)==0 );
	
	bdd_disable_reorder();
	//bdd_enable_reorder();
	//bdd_autoreorder(BDD_REORDER_WIN2ITE);
	
	// group prime and unprimed copies of variables together
	// this is to aid the dynamic var reordering
	//for(unsigned i=0; i<propvars; i++) {
	//	bdd_intaddvarblock(2*i, 2*i+1, BDD_REORDER_FIXED);
	//}

	// consider each variable its own reorder block
	for(unsigned i=0; i<propvars; i++) {
		bdd_intaddvarblock(i, i, BDD_REORDER_FIXED);
	}
	
	// set up our error handler
	bdd_error_hook(errorhandler);
		
	// setup gbc hook for cleaner output
	bdd_gbc_hook(NULL);

	// install variable print handler for dot output
	bdd_strm_hook(printhandler);	
		
	// set of vars to quantify on
/*	varset = bddtrue;
	for(unsigned i=0; i<propvars; i++) {
		// effectively the unprimed (first) copy of vars
		varset &= bdd_ithvar(2*i);
	}*/
	
	// pairs of vars to substitute
/*	substPairs = bdd_newpair();
	for(unsigned i=0; i<propvars; i++) {
		// add subst pair
		// primed var (second copy) is renamed to unprimed (first copy)
		bdd_setpair(substPairs, 2*i+1, 2*i);
	}
	
	build_d10();*/
}

// cleanup routine
void cleanup(void) {
	assert(bdd_isrunning());
	
/*	bdd_freepair(substPairs);
	substPairs = NULL;
	
	varset = bddfalse;
	d10 = bddfalse;*/
	
	bdd_clrvarblocks();
	
	bdd_done();
	
	// reset flags
	timedOut = false;
	outOfMem = false;
}

void print_bddstats(void) {
	using namespace std;
	
	bddStat stats;
	
	bdd_stats(&stats);
	
	cout << "Total number of nodes produced: " << stats.produced << endl;
	cout << "Currently allocated nodes:" << stats.nodenum << endl;
	cout << "User def max number of nodes: " << stats.maxnodenum << endl;
	cout << "Currently free nodes: " << stats.freenodes << endl;
	cout << "User def min number of free nodes: " << stats.minfreenodes << endl;
	cout << "Defined variables: " << stats.varnum << endl;
	cout << "Cache size in entries: " << stats.cachesize << endl;
	cout << "Number of garbage collections run: " << stats.gbcnum << endl;
	cout << endl;
}

//void check_reorder(void) {}

/*
void check_reorder(void) {
	int nodes = bdd_getallocnum();
	int used = bdd_getnodenum();
	double ratio = (double)used/(double)nodes;
	if(ratio > 0.01) {
		std::cerr << "reorder\n";
		bdd_reorder(BDD_REORDER_WIN2);
	}
}
*/

// // // returns d^1(phi) 
// bdd dilate(const bdd & phi) {
// 	//return bdd_replace( bdd_relprod( phi, d1, varset ), substPairs ) | phi;
// 	return bdd_replace( bdd_relprod( phi, d10, varset ), substPairs );
// }


// direct BDD algorithm for dilations
// pending evaluation for further use

#include <ext/hash_map>
//typedef __gnu_cxx::hash_map<int,bdd> DilationHashmap;

namespace __gnu_cxx {
	template<> struct hash<bdd> {
	public:
		size_t operator()(const bdd & x) const {
			return x.id();
		}
	};
}

typedef __gnu_cxx::hash_map<bdd,bdd> DilationHashmap;

static bdd dilate_ng(const bdd & A, DilationHashmap & dm) {
	if(A==bddtrue || A==bddfalse) {
		return A;
	}

	bdd high = bdd_high(A);
	if(high==bddtrue) {
		return bddtrue;
	}

	bdd low = bdd_low(A);
	if(low==bddtrue) {
		return bddtrue;
	}

// 	int id = A.id();
//  	DilationHashmap::const_iterator i = dm.find(id);
//  	if(i!=dm.end()) {
//  		return i->second;
//  	}

 	DilationHashmap::const_iterator i = dm.find(A);
 	if(i!=dm.end()) {
 		return i->second;
 	}

	// bdd_ite(condition, high, low)
	bdd result = bdd_ite(bdd_ithvar(bdd_var(A)), (dilate_ng(high, dm) | low), (dilate_ng(low, dm) | high));
	dm[A] = result;
	return result;
}

bdd dilate(const bdd & phi) {
	DilationHashmap dilationHashmap;
	return dilate_ng(phi, dilationHashmap);
}


