#include "foformula.h"
#include "operators/deltas.h"
#include "propformula.h"
#include <vector>
// auto_ptr
#include <memory>
#include <stack>
#include <cassert>
#include <string>
#include <sstream>
#include <map>
#include <set>

//-------------------------------------------------------------------------------------------------
class Valuation : public std::vector<unsigned> {
public:
	Valuation(unsigned v, const Signature & s) : std::vector<unsigned>(v,0), signature(s) {}
	bool next(void) {
		for(Valuation::iterator i=begin(); i!=end(); ++i) {
			++(*i);
			if(*i==signature.dsize) {
				*i = 0;
			} else {
				return true;
			}
		}
		return false;
	}
private:
	const Signature & signature;
};
//-------------------------------------------------------------------------------------------------
class Formula {
public:
	virtual bdd translate(const Signature & s, const Valuation & v) const = 0;
	virtual ~Formula() {}
};
//-------------------------------------------------------------------------------------------------
class SingleFormulaContainer: public Formula {
public:
	SingleFormulaContainer(const Formula * frm) : f(frm) {}
	virtual ~SingleFormulaContainer() {}
protected:
	const std::auto_ptr<const Formula> f;
};
//-------------------------------------------------------------------------------------------------
class TwoFormulaContainer: public Formula {
public:
	TwoFormulaContainer(const Formula * l, const Formula * r) : left(l), right(r) {}
	virtual ~TwoFormulaContainer() {}
protected:
	const std::auto_ptr<const Formula> left;
	const std::auto_ptr<const Formula> right;
	
};
//-------------------------------------------------------------------------------------------------
class Predicate: public Formula {
public:
	Predicate(unsigned p, const std::vector<unsigned> & a): predicate(p), args(a) {}
	virtual ~Predicate() {}
	virtual bdd translate(const Signature & s, const Valuation & v) const {
		unsigned factor = 1;
		unsigned var = 0;
		for(std::vector<unsigned>::const_iterator i=args.begin(); i!=args.end(); ++i) {
			var += v[*i] * factor;
			factor *= s.dsize;
		}
		var *= signature.predicates;
		var += predicate;
		return bdd_ithvar(var);
	}

private:
	unsigned predicate;
	const std::vector<unsigned> args;
};
//-------------------------------------------------------------------------------------------------
class Negation: public SingleFormulaContainer {
public:
	Negation(const Formula * frm) : SingleFormulaContainer(frm) {}
	virtual ~Negation() {}
	virtual bdd translate(const Signature & s, const Valuation & v) const {
		return !(f->translate(s,v));
	}
};
//-------------------------------------------------------------------------------------------------
class Disjunction: public TwoFormulaContainer {
public:
	Disjunction(const Formula * l, const Formula * r) : TwoFormulaContainer(l,r) {}
	virtual ~Disjunction() {}
	virtual bdd translate(const Signature & s, const Valuation & v) const {
		return (left->translate(s,v)) | (right->translate(s,v));
	}
};
//-------------------------------------------------------------------------------------------------
class Conjunction: public TwoFormulaContainer {
public:
	Conjunction(const Formula * l, const Formula * r) : TwoFormulaContainer(l,r) {}
	virtual ~Conjunction() {}
	virtual bdd translate(const Signature & s, const Valuation & v) const {
		return (left->translate(s,v)) & (right->translate(s,v));
	}
};
//-------------------------------------------------------------------------------------------------
class Implication: public TwoFormulaContainer {
public:
	Implication(const Formula * l, const Formula * r) : TwoFormulaContainer(l,r) {}
	virtual ~Implication() {}
	virtual bdd translate(const Signature & s, const Valuation & v) const {
		return (left->translate(s,v)) >> (right->translate(s,v));
	}
};
//-------------------------------------------------------------------------------------------------
class Biimplication: public TwoFormulaContainer {
public:
	Biimplication(const Formula * l, const Formula * r) : TwoFormulaContainer(l,r) {}
	virtual ~Biimplication() {}
	virtual bdd translate(const Signature & s, const Valuation & v) const {
		return bdd_biimp(left->translate(s,v), right->translate(s,v));
	}
};
//-------------------------------------------------------------------------------------------------
static bdd translate(unsigned variables, const Signature & s, const Formula * f) {
	Valuation v(variables, signature);
	bdd retval = bddtrue;

	do {
		retval &= f->translate(s, v);
	} while(v.next());
		
	return retval;
}
//-------------------------------------------------------------------------------------------------
static inline Formula * pop(std::stack<Formula *> & S) {
	Formula * f = S.top();
	S.pop();
	return f;
}
//-------------------------------------------------------------------------------------------------
static Formula * parseFormula(std::istream & in) {
	std::stack<Formula *> S;
	std::string temp;
	
	while (in >> temp) {
		if(temp.at(0)=='P') {
			unsigned p = (unsigned) (temp.at(1) - '0');
			assert(temp.size()==signature.arity+2);
			std::vector<unsigned> args(signature.arity, 0);
			for(unsigned i=0; i<signature.arity; i++) {
				args[i] = (unsigned)(temp.at(i+2) - '0');
			}
			S.push(new Predicate(p,args));
			continue;
		}
		
		if(temp=="not") {
			S.push(new Negation(pop(S)));
			continue;
		}
		
		if(temp=="and") {
			Formula * r = pop(S);
			Formula * l = pop(S);
			S.push(new Conjunction(l,r));
			continue;
		}
		
		if(temp=="iff") {
			Formula * r = pop(S);
			Formula * l = pop(S);
			S.push(new Biimplication(l,r));
			continue;
		}
		
		if(temp=="if") {
			Formula * r = pop(S);
			Formula * l = pop(S);
			S.push(new Implication(l,r));
			continue;
		}
		
		if(temp=="or") {
			Formula * r = pop(S);
			Formula * l = pop(S);
			S.push(new Disjunction(l,r));
			continue;
		}
		
		if(temp=="and") {
			Formula * r = pop(S);
			Formula * l = pop(S);
			S.push(new Conjunction(l,r));
			continue;
		}
		
		if(temp=="endformula") {
			break;
		}
		
		assert(1==0);
	}

	if(S.size()==1) {
		return S.top();
	} else {
		while(!S.empty()) {
			Formula * f = pop(S);
			delete f;
		}
		return NULL;
	}
}
//-------------------------------------------------------------------------------------------------
static KSet parseProfile(std::istream & in, unsigned kbs) {
	KSet profile;
	std::string temp;
	unsigned variables;
	
	
	for(unsigned i=0; i<kbs; ++i) {
		in >> temp;
		std::istringstream istr(temp);
		istr >> variables;
		assert(!istr.fail());
		profile.push_back(translate(variables, signature, parseFormula(in)));
	}
	
	return profile;
}
//-------------------------------------------------------------------------------------------------
void run_fo_input(unsigned kbs, unsigned op) {
	propvars = signature.predicates;
	for(unsigned i=0; i<signature.arity; i++) {
		propvars *= signature.dsize;
	}
	init();
	
	KSet profile = parseProfile(std::cin, kbs);
	bdd m = (*(ops[op]))(profile, bddtrue);
	
	//bdd_reorder_verbose(2);
//	bdd_reorder(BDD_REORDER_WIN3ITE);
	bdd_allsat(m, allsatPrintHandler);
// 	std::cout << bdddot << m << std::endl;

	CNF * f = buildCNF(m);
	std::cout << *f;
	delete f;
	
	cleanup();
}
//-------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------
/*typedef std::set<int> Conjunct;
// type for propositional DNF
typedef std::set<Conjunct> DNF;
DNF * buildDNF(bdd f);
// output a DNF 
std::ostream& operator<<(std::ostream& out, const DNF & f);
//----------------------------------------------------------------------------------------
// returns true if a variable exists in both polarities in a clause
// or if there are no variables in the clause
class TrivialClause {
public:
	inline bool operator()(const Conjunct & c) const {
		if(c.size()==0) {
			return true;
		}
		for(Clause::const_iterator i=c.begin(); i!=c.end(); ++i) {
			if(c.find(-(*i))!=c.end()) {
				return true;
			}
		}
		return false;
	}
};
//----------------------------------------------------------------------------------------
class Subsumes {
public:
	inline Subsumes(const DNF & F) : f(F) {}
	
	inline bool operator()(const Clause & c) const {
		for(CNF::const_iterator i=f.begin(); i!=f.end(); ++i) {
			if(c!=*i && std::includes(i->begin(), i->end(), c.begin(), c.end())) {
				return true;
			}
		}
		return false;
	}
	
private:
	const DNF & f;

};
//----------------------------------------------------------------------------------------
static void _buildDNF(bdd f, std::map<int,DNF> & cache) {
	if( f==bddtrue || f==bddfalse || cache.find(f.id())!=cache.end() ) {
		return;
	}
	
	cache[f.id()] = DNF();
	DNF & retval = cache[f.id()];
	
	bdd HIGH = bdd_high(f);
	bdd LOW = bdd_low(f);
	
	_buildDNF(HIGH, cache);
	_buildDNF(LOW, cache);
	
	const DNF & high = cache[HIGH.id()];
	const DNF & low = cache[LOW.id()];
	//int var = bdd_var(f)/2+1;
	int var = bdd_var(f)+1;
	
	// based on decomposing the ITE form var->HIGH,LOW to the DNF
	// (-var&LOW) | (var&HIGH)
	// and further decomposing based on DNF(HIGH), DNF(LOW)
	
	if(HIGH!=bddfalse) {
		if(HIGH==bddtrue) {
			Conjunct c;
			c.insert(var);
			retval.insert(c);
		} else {
			for(DNF::const_iterator i=high.begin(); i!=high.end(); ++i) {
				Conjunct c(*i);
				c.insert(var);
				retval.insert(c);
			}
		}
	}
		
	if(LOW!=bddfalse) {
		if(LOW==bddtrue) {
			Conjunct c;
			c.insert(-var);
			retval.insert(c);
		} else {
			for(DNF::const_iterator i=low.begin(); i!=low.end(); ++i) {
				Conjunct c(*i);
				c.insert(-var);
				retval.insert(c);
			}
		}
	}
		
	DNF tmp;
	
	// remove empty clauses and clauses that contain both polarities of a variable
	std::remove_copy_if(retval.begin(), retval.end(), std::inserter(tmp, tmp.begin()), TrivialClause());
	retval.swap(tmp);
	tmp.clear();
	
	// remove clauses that are subsumed by others
	std::remove_copy_if(retval.begin(), retval.end(), std::inserter(tmp, tmp.begin()), Subsumes(retval));
	retval.swap(tmp);
}
//----------------------------------------------------------------------------------------
DNF * buildDNF(bdd f) {
	std::map<int,DNF> cache;
	_buildDNF(f, cache);
	// f.id() points to the CNF version of f now
	// return dynamic copy
	return new DNF(cache[f.id()]);
}*/
