#include "deltas.h"

// enumerates sorted distance vectors in the G_max sense
class VecEnum_max {
	private:
		// stores information about distances from kbs
		// d[0] is the number of kbs at distance one in the distance vector
		// d[1] is the number of kbs at distance two ...
		// the sum of these numbers cannot be more than the number of kbs
		// the length of this is the maximum distance from any kb,
		// that is the number of propvars in the language
		IVec d;
		
		// actual distance vector
		IVec dv;
		unsigned kbs;
		
	public:
		// n represents the maximum distance from a single kb
		// k is the number of kbs
		VecEnum_max(unsigned n, unsigned k): d(n, 0), kbs(k) {}
		
		
		// O(propvars), O(kbs)
		void next(void) {
			// this loop works like an counter in base-max
			// but actually cycles when the sum of the digits 
			// is greater than max
			for(IVec::iterator ptr=d.begin(); ptr!=d.end(); ++ptr) {
				if((unsigned)vsum(d)<kbs) {
					++(*ptr);
					break;
				} else {
					*ptr = 0;
				}
			}
			
			// the rest converts the above to a sorted distance vector
			dv.clear();
			dv.reserve(kbs);
			
			for(int j=d.size()-1; j>=0; j--) {
				if(d.at(j)==0) {
					continue;
				}
				
				for(unsigned k=0; k<d.at(j); k++) {
					dv.push_back(j+1);
				}
			}
			
			while(dv.size()<kbs) {
				dv.push_back(0);
			}
		}
		
		const IVec * data(void) const {
			return &dv;
		}
};


bdd delta_G_max(const KSet &kset, const bdd &IC) {
	// check the trivial condition
	{
		bdd conj = conjunction(kset, IC);
		if(conj != bddfalse) {
			return conj;
		}
	}
	
	// 2-dimensional array keeping all dilations
	KSetSet dilation(kset.size());
	for(unsigned i=0; i<dilation.size(); i++) {
		// reserve space for up to propvars number of dilations for each formula
		dilation.at(i).reserve(propvars);
		// and initialise with the members of kset
		dilation.at(i).push_back( kset.at(i) );
	}

	bdd solution = bddfalse;
	
	// max distance is propvars
	//VecPerm_max vp(propvars, kset.size());
	VecEnum_max vm(propvars, kset.size());
	// skip 0-vector
	//vp.next();
	vm.next();
	
	for(unsigned j=1; !TERMINATE() && solution==bddfalse; ++j) {
		// dilate the top formulas
		for(KSetSet::iterator dptr=dilation.begin(); dptr!=dilation.end(); ++dptr) {
			dptr->push_back( dilate( *(dptr->end()-1) ) );
		}
		dilations++;

		// check that the conjunction of largest dilations is consistent, otherwise skip
		// whole set of permutations
		bool skip = false;
		{
			bdd conj = IC;
			for(KSetSet::const_iterator dptr=dilation.begin(); conj!=bddfalse && dptr!=dilation.end(); ++dptr) {
				conj &= *(dptr->end()-1);
			}
			skip = (conj==bddfalse);
		}
		
		if(skip) {
			// advance vector enumerator until it points to 
			// something that isn't below the maximal conjunction just tested
			do {
				vm.next();
			} while( vmax( *vm.data() )<=j );
			continue;
		}
		
		// enumerate permutations of distance vectors
		// until the maximum element grows beyond j+1 (no of dilations)
		do {
			// permutation of dist vector
			IVec p = *vm.data();
			// in order to start the permutation loop from the lex-minimum,
			// we need to sort in non-decreasing order
			// since we know that the original is non-increasing, reversing is enough
			std::reverse(p.begin(), p.end());
			
			do {
				// calculate conjunction with ICs
				bdd conj = IC;
				for(unsigned i=0; conj!=bddfalse && i<dilation.size(); ++i) {
					conj &= dilation.at(i).at(p.at(i));
				}
				conjunctions++;
				
				if(conj!=bddfalse) {
					solution |= conj;
				}

			} while ( !TERMINATE() && std::next_permutation(p.begin(), p.end()) );
		
			vm.next();
			
		} while( !TERMINATE() && solution==bddfalse && vmax( *vm.data() )<=j );
	}
	
	return solution;
}


// // enumerates permutations of VecEnum_lex in increasing order
// class VecPerm_max {
// 	private:
// 		VecEnum_max ve;
// 		IVec result;
// 	
// 	public:
// 		// n and m are as in VecEnum
// 		VecPerm_max(unsigned n, unsigned m) : ve(n, m), result(m, 0) {}
// 		
// 		const IVec * data(void) const {
// 			return &result;
// 		}
// 		
// 		// O(kbs), O(propvars)
// 		// returns false if next distance vector is required
// 		bool next(void) {
// 			// take next permutation and cycle ve if necessary
// 			// O(kbs)
// 			bool retval = std::next_permutation(result.begin(), result.end());
// 			if(!retval) {
// 				// O(kbs) O(propvars)
// 				ve.next();
// 				result = *(ve.data());
// 				// O(kbs)
// 				// in order to start the permutation loop from the lex-minimum,
// 				// we need to sort in non-decreasing order
// 				// since we know that the original is non-increasing, reversing is enough
// 				std::reverse(result.begin(), result.end());
// 			}
// 			return retval;
// 		}
// };
