#include "propformula.h"
#include "operators/deltas.h"
#include "dilations.h"

#include <sstream>
#include <string>
#include <stack>
#include <map>
#include <algorithm>
#include <iterator>

//----------------------------------------------------------------------------------------
static inline bdd pop(std::stack<bdd> & S) {
	bdd b = S.top();
	S.pop();
	return b;
}
//----------------------------------------------------------------------------------------
std::istream& operator>>(std::istream& in, bdd & formula) {
	std::stack<bdd> S;
	std::string temp;
	
	while (in >> temp) {
		std::istringstream istr(temp);
		int i;
		
		if (istr >> i) {
			S.push(bdd_ithvar(i));
			continue;
		}
		
		if(temp=="not") {
			S.push(bdd_not(pop(S)));
			continue;
		}
		
		if(temp=="iff") {
			bdd r = pop(S);
			bdd l = pop(S);
			S.push(bdd_biimp(l,r));
			continue;
		}
	
		if(temp=="and") {
			bdd r = pop(S);
			bdd l = pop(S);
			S.push(bdd_and(l,r));
			continue;
		}
	
		if(temp=="or") {
			bdd r = pop(S);
			bdd l = pop(S);
			S.push(bdd_or(l,r));
			continue;
		}
	
		if(temp=="if") {
			bdd r = pop(S);
			bdd l = pop(S);
			S.push(bdd_imp(l,r));
			continue;
		}
		
		if(temp=="endformula") {
			break;
		}
		
		assert(1==0);
	}

	assert(S.size()==1);
	formula = pop(S);
	return in;
}
//----------------------------------------------------------------------------------------
// returns true if a variable exists in both polarities in a clause
// or if there are no variables in the clause
class TrivialClause {
public:
	inline bool operator()(const Clause & c) const {
		if(c.size()==0) {
			return true;
		}
		for(Clause::const_iterator i=c.begin(); i!=c.end(); ++i) {
			if(c.find(-(*i))!=c.end()) {
				return true;
			}
		}
		return false;
	}
};
//----------------------------------------------------------------------------------------
class Subsumed {
public:
	inline Subsumed(const CNF & F) : f(F) {}
	
	inline bool operator()(const Clause & c) const {
		for(CNF::const_iterator i=f.begin(); i!=f.end(); ++i) {
			if(c!=*i && std::includes(c.begin(), c.end(), i->begin(), i->end())) {
				return true;
			}
		}
		return false;
	}
	
private:
	const CNF & f;

};
//----------------------------------------------------------------------------------------
static void _buildCNF(bdd f, std::map<int,CNF> & cache) {
	if( f==bddtrue || f==bddfalse || cache.find(f.id())!=cache.end() ) {
		return;
	}
	
	cache[f.id()] = CNF();
	CNF & retval = cache[f.id()];
	
	bdd HIGH = bdd_high(f);
	bdd LOW = bdd_low(f);
	
	_buildCNF(HIGH, cache);
	_buildCNF(LOW, cache);
	
	const CNF & high = cache[HIGH.id()];
	const CNF & low = cache[LOW.id()];
	int var = bdd_var(f)+1;
	
	
	// based on decomposing the ITE form var->HIGH,LOW to the CNF
	// (var | LOW) & (-var | HIGH) & (LOW | HIGH)
	// and further decomposing based on CNF(HIGH), CNF(LOW)
	
	if(HIGH!=bddtrue && LOW!=bddtrue) {
		if(HIGH==bddfalse && LOW==bddfalse) {
			retval.clear();
		} else if(HIGH==bddfalse) {
			retval.insert(low.begin(), low.end());
		} else if(LOW==bddfalse) {
			retval.insert(high.begin(), high.end());
		} else {
			for(CNF::const_iterator i=high.begin(); i!=high.end(); ++i) {
				for(CNF::const_iterator j=low.begin(); j!=low.end(); ++j) {
					Clause c;
					std::set_union(i->begin(), i->end(), j->begin(), j->end(), std::inserter(c, c.begin()));
					retval.insert(c);
				}
			}
		}
	}
	
	if(HIGH!=bddtrue) {
		if(HIGH==bddfalse) {
			Clause c;
			c.insert(-var);
			retval.insert(c);
		} else {
			for(CNF::const_iterator i=high.begin(); i!=high.end(); ++i) {
				Clause c(*i);
				c.insert(-var);
				retval.insert(c);
			}
		}
	}
		
	if(LOW!=bddtrue) {
		if(LOW==bddfalse) {
			Clause c;
			c.insert(var);
			retval.insert(c);
		} else {
			for(CNF::const_iterator i=low.begin(); i!=low.end(); ++i) {
				Clause c(*i);
				c.insert(var);
				retval.insert(c);
			}
		}
	}
		
	CNF tmp;
	
	// remove empty clauses and clauses that contain both polarities of a variable
	std::remove_copy_if(retval.begin(), retval.end(), std::inserter(tmp, tmp.begin()), TrivialClause());
	retval.swap(tmp);
	tmp.clear();
	
	// remove clauses that are subsumed by others
	std::remove_copy_if(retval.begin(), retval.end(), std::inserter(tmp, tmp.begin()), Subsumed(retval));
	retval.swap(tmp);
}
//----------------------------------------------------------------------------------------
CNF * buildCNF(bdd f) {
	std::map<int,CNF> cache;
	_buildCNF(f, cache);
	// f.id() points to the CNF version of f now
	// return dynamic copy
	return new CNF(cache[f.id()]);
}
//----------------------------------------------------------------------------------------
std::ostream& operator<<(std::ostream& out, const CNF & f) {
	out<<f.size() << '\n';
	for(CNF::const_iterator i=f.begin(); i!=f.end(); ++i) {
/*		for(Clause::const_iterator j=i->begin(); j!=i->end(); ++j) {
			out << *j << ' ';
		}
		out << '\n';*/
		int m = std::max(abs(*(i->begin())), abs(*(i->rbegin())));
		for(int j=1; j<=m; ++j) {
			if(i->find(j)!=i->end()) {
				out << ' ' << j << ' ';
			} else if(i->find(-j)!=i->end()) {
				out << -j << ' ';
			}
		}
		out << '\n';
	}
	return out;
}
//----------------------------------------------------------------------------------------
static void _printITE(bdd f, std::set<int> & printed) {
	if(f==bddtrue || f==bddfalse) {
		return;
	} 
	
	if(printed.find(f.id())!=printed.end()) {
		return;
	}

	printed.insert(f.id());
	_printITE(bdd_low(f), printed);
	_printITE(bdd_high(f), printed);
	std::cout << f.id() << ',' << bdd_var(f)+1 << ',' << bdd_high(f).id() << ',' << bdd_low(f).id() << '\n';
}

void printITE(bdd f) {
	std::set<int> printed;
	_printITE(f, printed);
}
//-------------------------------------------------------------------------------------------
void run_prop_input(unsigned kbs, unsigned op) {
	init();
	
	KSet profile;
	bdd IC;

	std::cin >> IC;

	for(unsigned i=0; i<kbs; ++i) {
		bdd phi;
		std::cin >> phi;
		profile.push_back(phi);
	}

/*	for(int i=0; i<kbs; ++i) {
		std::cout << '[' << std::endl;
		bdd_allsat(profile[i], allsatPrintHandler);
		std::cout << ']' << std::endl;
	}*/
	
	bdd m = (*(ops[op]))(profile, IC);

	//std::cout << printFormula(m) << std::endl;
//	std::cout << '[' << std::endl;
	bdd_allsat(m, allsatPrintHandler);
//	std::cout << ']' << std::endl;
	//std::cout << bdd_satcount(m) << std::endl;
//	printITE(m);
	//std::cout << bdd_nodecount(m) << std::endl; 
	
	//CNF * f = buildCNF(m);
	//std::cout << *f;
	//delete f;
	
	cleanup();
}
//----------------------------------------------------------------------------------------
