#include "randmerge.h"
#include "dilations.h"
#include "benchmark.h"
#include "operators/deltas.h"
#include "util/mt19937ar.h"

#define boundedRandom(b) ( genrand_int32() % (b) )
//-------------------------------------------------------------------------------------------------
// generates random formulas in CNF distributed along a KSet
class FormulaGenerator {
	private:
		unsigned kbSize;
		unsigned clausesPerFormula;
		unsigned varsPerClause;

		// will generate one 3-CNF formula with up to clausesPerFormula clauses
		bdd generateFormula(void) const;
		
		
	public:
		// no of knowledgebases, no of clauses per KB, no of variables per clause
		void setup(unsigned kbs, unsigned cl, unsigned vc);
		bool generate(unsigned seed, KSet * kset, bdd * IC) const;
		FormulaGenerator(unsigned kbs, unsigned cl, unsigned vc) { setup(kbs, cl, vc); }
};
//-------------------------------------------------------------------------------------------------
void FormulaGenerator::setup(unsigned kbs, unsigned cl, unsigned vc) {
	kbSize = kbs;
	clausesPerFormula = cl;
	varsPerClause = vc;
}
//-------------------------------------------------------------------------------------------------
bdd FormulaGenerator::generateFormula(void) const {
	bdd formula = bddtrue;
	
	// generate a formula
	for(unsigned ic=0; formula!=bddfalse && ic<clausesPerFormula; ic++) {
		bdd clause = bddfalse;

		// generate a clause
		for(unsigned iv=0; iv<varsPerClause; iv++) {
			int v = (int) (boundedRandom(propvars));
			
			if(genrand_real2()<0.5) {
				clause |= bdd_ithvar(v);
			} else {
				clause |= bdd_nithvar(v);
			}
		}
		
		// conjoin with formula
		formula &= clause;
	}

	return formula;
}
//-------------------------------------------------------------------------------------------------
bool FormulaGenerator::generate(unsigned seed, KSet * kset, bdd * IC) const {
	// setup random number generator
	init_genrand(seed);

	kset->clear();
	kset->reserve(kbSize);
	
	*IC = generateFormula();
	if(*IC==bddfalse) {
		//std::cerr << "IC\n";
		return false;
	}
	
	for(unsigned ik=0; ik<kbSize; ik++) {
		bdd formula = generateFormula();
		
		if(formula==bddfalse) {
			//std::cerr << "FO\n";
			return false;
		}
		
		kset->push_back(formula);
	}
	
	return true;
}
//-------------------------------------------------------------------------------------------
// test one operator, changing seed used until a valid profile has been found
void run_random_oneshot(unsigned kbs, unsigned clausesPerKB, unsigned * seed, int op) {
	Benchmark formBmk;
	Benchmark mergeBmk;
	FormulaGenerator fg(kbs, clausesPerKB, 3);
	bdd IC;
	KSet kset;
	bool found=false;

	while(true) {
		init();
		formBmk.start(false);
		// keep going until we get a profile that doesn't run out of memory
		found = fg.generate(*seed, &kset, &IC) && !outOfMem;
		formBmk.stop(kset, IC);
		
		if(found) {
			break;
		}
		
		kset.clear();
		IC = bddfalse;
		cleanup();
		*seed = genrand_int32();
	}

	// now benchmark merging
	mergeBmk.start(true);
	mergeBmk.stop( (*(ops[op])) (kset, IC) );
	
	// now that benchmarking is over, calculate consistency of profile+IC
	int inconsistent = (int) (outOfMem || conjunction(kset, IC)==bddfalse);
	
	// output
	std::cout << "P," << *seed << ',' << inconsistent << ',' << formBmk << ',';
	std::cout << "M," << *seed << ',' << op+1 << ',' << mergeBmk << '\n';
	
	// clear all bdd vars before calling cleanup
	// to avoid referencing trouble
	kset.clear();
	IC = bddfalse;
	cleanup();
}
//-------------------------------------------------------------------------------------------
// merge several times (repetitions), starting from seed, for an operator
void run_random(unsigned kbs, unsigned clausesPerKB, unsigned seed, unsigned repetitions, int op) {
	for(unsigned i=0; i<repetitions; i++) {
		run_random_oneshot(kbs, clausesPerKB, &seed, op);
		seed = genrand_int32();
	}
}
//-------------------------------------------------------------------------------------------------

